@@ -45,9 +45,12 @@
 #include "util.h"
 #include "drop_privs.h"
 
+#include <ykclient.h>
+
 #if HAVE_CR
-/* for yubikey_hex_decode and yubikey_hex_p */
+/* for yubikey_hex_encode */
 #include <yubikey.h>
+/* for yubikey pbkdf2*/
 #include <ykpbkdf2.h>
 #endif /* HAVE_CR */
 
@@ -97,7 +100,7 @@ enum key_mode {
 
 struct cfg
 {
-  int client_id;
+  unsigned int client_id;
   const char *client_key;
   int debug;
   int alwaysok;
@@ -110,11 +113,15 @@ struct cfg
   const char *urllist;
   const char *ldapserver;
   const char *ldap_uri;
+  const char *ldap_bind_user;
+  const char *ldap_bind_password;
+  const char *ldap_filter;
+  const char *ldap_cacertfile;
   const char *ldapdn;
   const char *user_attr;
   const char *yubi_attr;
   const char *yubi_attr_prefix;
-  int token_id_length;
+  unsigned int token_id_length;
   enum key_mode mode;
   const char *chalresp_path;
 };
@@ -124,84 +131,6 @@ struct cfg
 #endif
 #define DBG(x) if (cfg->debug) { D(x); }
 
-/*
- * This function will look for users name with valid user token id. It
- * will returns -2 if the user is unknown, -1 if the token do not match the user line, 0 for internal failure and 1 for success.
- *
- * File format is as follows:
- * <user-name>:<token_id>:<token_id>
- * <user-name>:<token_id>
- *
- */
-static int
-check_user_token (struct cfg *cfg,
-		  const char *authfile,
-		  const char *username,
-		  const char *otp_id)
-{
-  char buf[1024];
-  char *s_user, *s_token;
-  int retval = 0;
-  int fd;
-  struct stat st;
-  FILE *opwfile;
-
-  fd = open(authfile, O_RDONLY, 0);
-  if (fd < 0) {
-      DBG (("Cannot open file: %s (%s)", authfile, strerror(errno)));
-      return retval;
-  }
-
-  if (fstat(fd, &st) < 0) {
-      DBG (("Cannot stat file: %s (%s)", authfile, strerror(errno)));
-      close(fd);
-      return retval;
-  }
-
-  if (!S_ISREG(st.st_mode)) {
-      DBG (("%s is not a regular file", authfile));
-      close(fd);
-      return retval;
-  }
-
-  opwfile = fdopen(fd, "r");
-  if (opwfile == NULL) {
-      DBG (("fdopen: %s", strerror(errno)));
-      close(fd);
-      return retval;
-  }
-
-  retval = -2;
-  while (fgets (buf, 1024, opwfile))
-    {
-      if (buf[strlen (buf) - 1] == '\n')
-	buf[strlen (buf) - 1] = '\0';
-      DBG (("Authorization line: %s", buf));
-      s_user = strtok (buf, ":");
-      if (s_user && strcmp (username, s_user) == 0)
-	{
-	  DBG (("Matched user: %s", s_user));
-      retval = -1; //We found at least one line for the user
-	  do
-	    {
-	      s_token = strtok (NULL, ":");
-	      DBG (("Authorization token: %s", s_token));
-	      if (s_token && strcmp (otp_id, s_token) == 0)
-		{
-		  DBG (("Match user/token as %s/%s", username, otp_id));
-		  fclose (opwfile);
-		  return 1;
-		}
-	    }
-	  while (s_token != NULL);
-	}
-    }
-
-  fclose (opwfile);
-
-  return retval;
-}
-
 /*
  * Authorize authenticated OTP_ID for login as USERNAME using
  * AUTHFILE.  Return -2 if the user is unknown, -1 if the OTP_ID does not match,  0 on internal failures, otherwise success.
@@ -220,7 +149,7 @@ authorize_user_token (struct cfg *cfg,
          as an argument for this module.
        */
       DBG (("Using system-wide auth_file %s", cfg->auth_file));
-      retval = check_user_token (cfg, cfg->auth_file, username, otp_id);
+      retval = check_user_token (cfg->auth_file, username, otp_id, cfg->debug);
     }
   else
     {
@@ -249,7 +178,7 @@ authorize_user_token (struct cfg *cfg,
 	goto free_out;
       }
 
-      retval = check_user_token (cfg, userfile, username, otp_id);
+      retval = check_user_token (userfile, username, otp_id, cfg->debug);
 
       if(pam_modutil_regain_priv(pamh, &privs)) {
         DBG (("could not restore privileges"));
@@ -268,8 +197,10 @@ free_out:
  * This function will look in ldap id the token correspond to the
  * requested user. It will returns 0 for failure and 1 for success.
  *
- * For the moment ldaps is not supported. ldap serve can be on a
- * remote host.
+ * ldaps is only supported for ldap_uri based connections.
+ * ldap_cacertfile usually needs to be set for this to work.
+ *
+ * ldap serve can be on a remote host.
  *
  * You need the following parameters in you pam config:
  * ldapserver=  OR ldap_uri=
@@ -277,6 +208,9 @@ free_out:
  * user_attr=
  * yubi_attr=
  *
+ * If using ldap_uri, you can specify multiple failover hosts
+ * eg.
+ * ldap_uri=ldaps://host1.fqdn.example.com,ldaps://host2.fqdn.example.com
  */
 static int
 authorize_user_token_ldap (struct cfg *cfg,
@@ -284,8 +218,9 @@ authorize_user_token_ldap (struct cfg *cfg,
 			   const char *token_id)
 {
   int retval = 0;
-  int protocol;
 #ifdef HAVE_LIBLDAP
+  /* LDAPv2 is historical -- RFC3494. */
+  int protocol = LDAP_VERSION3;
   int yubi_attr_prefix_len = 0;
   LDAP *ld = NULL;
   LDAPMessage *result = NULL, *e;
@@ -296,21 +231,18 @@ authorize_user_token_ldap (struct cfg *cfg,
   struct berval **vals;
   int i, rc;
 
+  char *filter = NULL;
   char *find = NULL;
+  int scope = LDAP_SCOPE_BASE;
 #endif
   DBG(("called"));
 #ifdef HAVE_LIBLDAP
-
-  if (cfg->user_attr == NULL) {
-    DBG (("Trying to look up user to YubiKey mapping in LDAP, but user_attr not set!"));
-    return 0;
-  }
   if (cfg->yubi_attr == NULL) {
     DBG (("Trying to look up user to YubiKey mapping in LDAP, but yubi_attr not set!"));
     return 0;
   }
-  if (cfg->ldapdn == NULL) {
-    DBG (("Trying to look up user to YubiKey mapping in LDAP, but ldapdn not set!"));
+  if (cfg->user_attr && cfg->ldapdn == NULL) {
+    DBG (("Trying to look up user to YubiKey mapping in LDAP, user_attr set but ldapdn not set!"));
     return 0;
   }
 
@@ -320,7 +252,7 @@ authorize_user_token_ldap (struct cfg *cfg,
       rc = ldap_initialize (&ld, cfg->ldap_uri);
       if (rc != LDAP_SUCCESS)
 	{
-	  DBG (("ldap_init: %s", ldap_err2string (rc)));
+	  DBG (("ldap_initialize: %s", ldap_err2string (rc)));
 	  retval = 0;
 	  goto done;
 	}
@@ -335,12 +267,21 @@ authorize_user_token_ldap (struct cfg *cfg,
 	}
     }
 
-  /* LDAPv2 is historical -- RFC3494. */
-  protocol = LDAP_VERSION3;
+  ldap_set_option(ld, LDAP_OPT_REFERRALS, LDAP_OPT_OFF);
   ldap_set_option (ld, LDAP_OPT_PROTOCOL_VERSION, &protocol);
 
+  if (cfg->ldap_uri && cfg->ldap_cacertfile) {
+    /* Set CA CERTFILE.  This makes ldaps work when using ldap_uri */
+    ldap_set_option (0, LDAP_OPT_X_TLS_CACERTFILE, cfg->ldap_cacertfile);
+  }
   /* Bind anonymously to the LDAP server. */
-  rc = ldap_simple_bind_s (ld, NULL, NULL);
+  if (cfg->ldap_bind_user && cfg->ldap_bind_password) {
+    DBG (("try bind with: %s:[%s]", cfg->ldap_bind_user, cfg->ldap_bind_password));
+    rc = ldap_simple_bind_s (ld, cfg->ldap_bind_user, cfg->ldap_bind_password);
+  } else {
+    DBG (("try bind anonymous"));
+    rc = ldap_simple_bind_s (ld, NULL, NULL);
+  }
   if (rc != LDAP_SUCCESS)
     {
       DBG (("ldap_simple_bind_s: %s", ldap_err2string (rc)));
@@ -349,22 +290,30 @@ authorize_user_token_ldap (struct cfg *cfg,
     }
 
   /* Allocation of memory for search strings depending on input size */
-  i = (strlen(cfg->user_attr) + strlen(cfg->ldapdn) + strlen(user) + 3) * sizeof(char);
-  if ((find = malloc(i)) == NULL) {
-    DBG (("Failed allocating %i bytes", i));
-    retval = 0;
-    goto done;
+  if (cfg->user_attr && cfg->yubi_attr && cfg->ldapdn) {
+    i = (strlen(cfg->user_attr) + strlen(cfg->ldapdn) + strlen(user) + 3) * sizeof(char);
+    if ((find = malloc(i)) == NULL) {
+      DBG (("Failed allocating %i bytes", i));
+      retval = 0;
+      goto done;
+    }
+    sprintf (find, "%s=%s,%s", cfg->user_attr, user, cfg->ldapdn);
+    filter = NULL;
+  } else if (cfg->ldapdn) {
+    find = strdup(cfg->ldapdn); /* allow free later */
+  }
+  if (cfg->ldap_filter) {
+    filter = filter_printf(cfg->ldap_filter, user);
+    scope = LDAP_SCOPE_SUBTREE;
   }
-
-  sprintf (find, "%s=%s,%s", cfg->user_attr, user, cfg->ldapdn);
-
   attrs[0] = (char *) cfg->yubi_attr;
 
-  DBG(("LDAP : look up object '%s', ask for attribute '%s'", find, cfg->yubi_attr));
+  DBG(("LDAP : look up object base='%s' filter='%s', ask for attribute '%s'", find,
+      filter ? filter:"(null)", cfg->yubi_attr));
 
   /* Search for the entry. */
-  if ((rc = ldap_search_ext_s (ld, find, LDAP_SCOPE_BASE,
-			       NULL, attrs, 0, NULL, NULL, LDAP_NO_LIMIT,
+  if ((rc = ldap_search_ext_s (ld, find, scope,
+			       filter, attrs, 0, NULL, NULL, LDAP_NO_LIMIT,
 			       LDAP_NO_LIMIT, &result)) != LDAP_SUCCESS)
     {
       DBG (("ldap_search_ext_s: %s", ldap_err2string (rc)));
@@ -388,16 +337,16 @@ authorize_user_token_ldap (struct cfg *cfg,
 	{
 	  if ((vals = ldap_get_values_len (ld, e, a)) != NULL)
 	    {
-	      DBG(("LDAP : Found %i values - checking if any of them match '%s%s'",
-		   ldap_count_values_len(vals),
-		   cfg->yubi_attr_prefix ? cfg->yubi_attr_prefix : "",
-		   token_id));
-
 	      yubi_attr_prefix_len = cfg->yubi_attr_prefix ? strlen(cfg->yubi_attr_prefix) : 0;
 
 	      /* Compare each value for the attribute against the token id. */
 	      for (i = 0; vals[i] != NULL; i++)
 		{
+	          DBG(("LDAP : Found %i values - checking if any of them match '%s:%s:%s'",
+		       ldap_count_values_len(vals),
+		       vals[i]->bv_val,
+		       cfg->yubi_attr_prefix ? cfg->yubi_attr_prefix : "", token_id));
+
 		  /* Only values containing this prefix are considered. */
 		  if ((!cfg->yubi_attr_prefix || !strncmp (cfg->yubi_attr_prefix, vals[i]->bv_val, yubi_attr_prefix_len)))
 		    {
@@ -425,6 +374,8 @@ authorize_user_token_ldap (struct cfg *cfg,
   /* free memory allocated for search strings */
   if (find != NULL)
     free(find);
+  if (filter != NULL)
+    free(filter);
 
 #else
   DBG (("Trying to use LDAP, but this function is not compiled in pam_yubico!!"));
@@ -458,7 +409,13 @@ display_error(pam_handle_t *pamh, const char *message) {
     return retval;
   }
 
-  D(("conv returned: '%s'", resp->resp));
+  if (resp)
+    {
+      D(("conv returned: '%s'", resp->resp));
+      if (resp->resp)
+        free (resp->resp);
+      free (resp);
+    }
   return retval;
 }
 #endif /* HAVE_CR */
@@ -568,7 +525,7 @@ do_challenge_response(pam_handle_t *pamh, struct cfg *cfg, const char *username)
    */
 
   yubikey_hex_encode(response_hex, buf, response_len);
-  if(state.salt_len > 0) { // the expected response has gone through pbkdf2
+  if(state.salt_len > 0) { /* the expected response has gone through pbkdf2 */
     YK_PRF_METHOD prf_method = {20, yk_hmac_sha1};
     yk_pbkdf2(response_hex, (unsigned char*)state.salt, state.salt_len, state.iterations,
         (unsigned char*)buf, response_len, &prf_method);
@@ -709,7 +666,7 @@ parse_cfg (int flags, int argc, const char **argv, struct cfg *cfg)
   int i;
 
   memset (cfg, 0, sizeof(struct cfg));
-  cfg->client_id = -1;
+  cfg->client_id = 0;
   cfg->token_id_length = DEFAULT_TOKEN_ID_LEN;
   cfg->mode = CLIENT;
 
@@ -741,6 +698,14 @@ parse_cfg (int flags, int argc, const char **argv, struct cfg *cfg)
 	cfg->ldapserver = argv[i] + 11;
       if (strncmp (argv[i], "ldap_uri=", 9) == 0)
 	cfg->ldap_uri = argv[i] + 9;
+      if (strncmp (argv[i], "ldap_bind_user=", 15) == 0)
+	cfg->ldap_bind_user = argv[i] + 15;
+      if (strncmp (argv[i], "ldap_bind_password=", 19) == 0)
+	cfg->ldap_bind_password = argv[i] + 19;
+      if (strncmp (argv[i], "ldap_filter=", 12) == 0)
+	cfg->ldap_filter = argv[i] + 12;
+      if (strncmp (argv[i], "ldap_cacertfile=", 16) == 0)
+        cfg->ldap_cacertfile = argv[i] + 16;
       if (strncmp (argv[i], "ldapdn=", 7) == 0)
 	cfg->ldapdn = argv[i] + 7;
       if (strncmp (argv[i], "user_attr=", 10) == 0)
@@ -750,7 +715,7 @@ parse_cfg (int flags, int argc, const char **argv, struct cfg *cfg)
       if (strncmp (argv[i], "yubi_attr_prefix=", 17) == 0)
 	cfg->yubi_attr_prefix = argv[i] + 17;
       if (strncmp (argv[i], "token_id_length=", 16) == 0)
-	sscanf (argv[i], "token_id_length=%d", &cfg->token_id_length);
+	sscanf (argv[i], "token_id_length=%u", &cfg->token_id_length);
       if (strcmp (argv[i], "mode=challenge-response") == 0)
 	cfg->mode = CHRESP;
       if (strcmp (argv[i], "mode=client") == 0)
@@ -765,7 +730,7 @@ parse_cfg (int flags, int argc, const char **argv, struct cfg *cfg)
       D (("flags %d argc %d", flags, argc));
       for (i = 0; i < argc; i++)
 	D (("argv[%d]=%s", i, argv[i]));
-      D (("id=%d", cfg->client_id));
+      D (("id=%u", cfg->client_id));
       D (("key=%s", cfg->client_key ? cfg->client_key : "(null)"));
       D (("debug=%d", cfg->debug));
       D (("alwaysok=%d", cfg->alwaysok));
@@ -775,6 +740,10 @@ parse_cfg (int flags, int argc, const char **argv, struct cfg *cfg)
       D (("authfile=%s", cfg->auth_file ? cfg->auth_file : "(null)"));
       D (("ldapserver=%s", cfg->ldapserver ? cfg->ldapserver : "(null)"));
       D (("ldap_uri=%s", cfg->ldap_uri ? cfg->ldap_uri : "(null)"));
+      D (("ldap_bind_user=%s", cfg->ldap_bind_user ? cfg->ldap_bind_user : "(null)"));
+      D (("ldap_bind_password=%s", cfg->ldap_bind_password ? cfg->ldap_bind_password : "(null)"));
+      D (("ldap_filter=%s", cfg->ldap_filter ? cfg->ldap_filter : "(null)"));
+      D (("ldap_cacertfile=%s", cfg->ldap_cacertfile ? cfg->ldap_cacertfile : "(null)"));
       D (("ldapdn=%s", cfg->ldapdn ? cfg->ldapdn : "(null)"));
       D (("user_attr=%s", cfg->user_attr ? cfg->user_attr : "(null)"));
       D (("yubi_attr=%s", cfg->yubi_attr ? cfg->yubi_attr : "(null)"));
@@ -797,13 +766,13 @@ pam_sm_authenticate (pam_handle_t * pamh,
   const char *password = NULL;
   char otp[MAX_TOKEN_ID_LEN + TOKEN_OTP_LEN + 1] = { 0 };
   char otp_id[MAX_TOKEN_ID_LEN + 1] = { 0 };
-  int password_len = 0;
+  size_t password_len = 0;
   int skip_bytes = 0;
   int valid_token = 0;
   struct pam_conv *conv;
   const struct pam_message *pmsg[1];
   struct pam_message msg[1];
-  struct pam_response *resp;
+  struct pam_response *resp = NULL;
   int nargs = 1;
   ykclient_t *ykc = NULL;
   struct cfg cfg_st;
@@ -811,9 +780,12 @@ pam_sm_authenticate (pam_handle_t * pamh,
   size_t templates = 0;
   char *urls[10];
   char *tmpurl = NULL;
+  char *onlypasswd = NULL;
 
   parse_cfg (flags, argc, argv, cfg);
 
+  DBG (("pam_yubico version: %s", VERSION));
+
   if (cfg->token_id_length > MAX_TOKEN_ID_LEN)
   {
     DBG (("configuration error: token_id_length too long. Maximum acceptable value : %d", MAX_TOKEN_ID_LEN));
@@ -858,6 +830,12 @@ pam_sm_authenticate (pam_handle_t * pamh,
       goto done;
     }
 
+  if(ykclient_global_init() != YKCLIENT_OK)
+    {
+      DBG (("Failed initializing ykclient library"));
+      retval = PAM_AUTHINFO_UNAVAIL;
+      goto done;
+    }
   rc = ykclient_init (&ykc);
   if (rc != YKCLIENT_OK)
     {
@@ -950,7 +928,6 @@ pam_sm_authenticate (pam_handle_t * pamh,
 	  }
       }
       msg[0].msg_style = cfg->verbose_otp ? PAM_PROMPT_ECHO_ON : PAM_PROMPT_ECHO_OFF;
-      resp = NULL;
 
       retval = conv->conv (nargs, pmsg, &resp, conv->appdata_ptr);
 
@@ -999,7 +976,7 @@ pam_sm_authenticate (pam_handle_t * pamh,
   /* user entered their system password followed by generated OTP? */
   if (password_len > TOKEN_OTP_LEN + cfg->token_id_length)
     {
-      char *onlypasswd = strdup (password);
+      onlypasswd = strdup (password);
 
       if (! onlypasswd) {
 	retval = PAM_BUF_ERR;
@@ -1012,7 +989,6 @@ pam_sm_authenticate (pam_handle_t * pamh,
 	    "setting item PAM_AUTHTOK"));
 
       retval = pam_set_item (pamh, PAM_AUTHTOK, onlypasswd);
-      free (onlypasswd);
       if (retval != PAM_SUCCESS)
 	{
 	  DBG (("set_item returned error: %s", pam_strerror (pamh, retval)));
@@ -1071,6 +1047,8 @@ pam_sm_authenticate (pam_handle_t * pamh,
     }
 
 done:
+  if (onlypasswd)
+    free(onlypasswd);
   if (templates > 0)
     {
       size_t i;
@@ -1082,14 +1060,24 @@ done:
   if (tmpurl)
     free(tmpurl);
   if (ykc)
-    ykclient_done (&ykc);
+    {
+      ykclient_done (&ykc);
+      ykclient_global_done();
+    }
   if (cfg->alwaysok && retval != PAM_SUCCESS)
     {
       DBG (("alwaysok needed (otherwise return with %d)", retval));
       retval = PAM_SUCCESS;
     }
   DBG (("done. [%s]", pam_strerror (pamh, retval)));
-  pam_set_data (pamh, "yubico_setcred_return", (void*) (intptr_t) retval, NULL);
+  pam_set_data (pamh, "yubico_setcred_return", (void*)(intptr_t)retval, NULL);
+
+  if (resp)
+    {
+      if (resp->resp)
+        free (resp->resp);
+      free (resp);
+    }
 
   return retval;
 }
@@ -1100,16 +1088,55 @@ pam_sm_setcred (pam_handle_t * pamh, int flags, int argc, const char **argv)
   return PAM_SUCCESS;
 }
 
+PAM_EXTERN int
+pam_sm_acct_mgmt(pam_handle_t *pamh, int flags, int argc, const char **argv)
+{
+  int retval;
+  int rc = pam_get_data(pamh, "yubico_setcred_return", (const void**)&retval);
+  if (rc == PAM_SUCCESS && retval == PAM_SUCCESS) {
+    D (("pam_sm_acct_mgmt returing PAM_SUCCESS"));
+    return PAM_SUCCESS;
+  }
+  D (("pam_sm_acct_mgmt returing PAM_AUTH_ERR:%d", rc));
+  return PAM_AUTH_ERR;
+}
+
+PAM_EXTERN int
+pam_sm_open_session(pam_handle_t *pamh, int flags,
+	int argc, const char *argv[])
+{
+
+  D(("pam_sm_open_session"));
+  return (PAM_SUCCESS);
+}
+
+PAM_EXTERN int
+pam_sm_close_session(pam_handle_t *pamh, int flags,
+	int argc, const char *argv[])
+{
+  D(("pam_sm_close_session"));
+  return (PAM_SUCCESS);
+}
+
+PAM_EXTERN int
+pam_sm_chauthtok(pam_handle_t *pamh, int flags,
+	int argc, const char *argv[])
+{
+  D(("pam_sm_chauthtok"));
+  return (PAM_SERVICE_ERR);
+}
+
+
 #ifdef PAM_STATIC
 
 struct pam_module _pam_yubico_modstruct = {
   "pam_yubico",
   pam_sm_authenticate,
   pam_sm_setcred,
-  NULL,
-  NULL,
-  NULL,
-  NULL
+  pam_sm_acct_mgmt,
+  pam_sm_open_session,
+  pam_sm_close_session,
+  pam_sm_chauthtok
 };
 
 #endif
