@@ -35,6 +35,9 @@
 #include <stdlib.h>
 #include <string.h>
 #include <sys/types.h>
+#include <sys/stat.h>
+#include <errno.h>
+#include <fcntl.h>
 #include <pwd.h>
 #include <unistd.h>
 
@@ -44,6 +47,9 @@
 /* for yubikey_hex_decode and yubikey_hex_p */
 #include <yubikey.h>
 #include <ykpbkdf2.h>
+
+#include <ykstatus.h>
+#include <ykdef.h>
 #endif /* HAVE_CR */
 
 int
@@ -83,6 +89,94 @@ get_user_cfgfile_path(const char *common_path, const char *filename, const char
   return 1;
 }
 
+
+/*
+ * This function will look for users name with valid user token id. It
+ * will returns -2 if the user is unknown, -1 if the token do not match the user line, 0 for internal failure and 1 for success.
+ *
+ * File format is as follows:
+ * <user-name>:<token_id>:<token_id>
+ * <user-name>:<token_id>
+ *
+ */
+int
+check_user_token (const char *authfile,
+		  const char *username,
+		  const char *otp_id,
+		  int verbose)
+{
+  char buf[1024];
+  char *s_user, *s_token;
+  int retval = 0;
+  int fd;
+  struct stat st;
+  FILE *opwfile;
+
+  fd = open(authfile, O_RDONLY, 0);
+  if (fd < 0) {
+      if(verbose)
+	  D (("Cannot open file: %s (%s)", authfile, strerror(errno)));
+      return retval;
+  }
+
+  if (fstat(fd, &st) < 0) {
+      if(verbose)
+	  D (("Cannot stat file: %s (%s)", authfile, strerror(errno)));
+      close(fd);
+      return retval;
+  }
+
+  if (!S_ISREG(st.st_mode)) {
+      if(verbose)
+	  D (("%s is not a regular file", authfile));
+      close(fd);
+      return retval;
+  }
+
+  opwfile = fdopen(fd, "r");
+  if (opwfile == NULL) {
+      if(verbose)
+	  D (("fdopen: %s", strerror(errno)));
+      close(fd);
+      return retval;
+  }
+
+  retval = -2;
+  while (fgets (buf, 1024, opwfile))
+    {
+      char *saveptr = NULL;
+      if (buf[strlen (buf) - 1] == '\n')
+	buf[strlen (buf) - 1] = '\0';
+      if(verbose)
+	  D (("Authorization line: %s", buf));
+      s_user = strtok_r (buf, ":", &saveptr);
+      if (s_user && strcmp (username, s_user) == 0)
+	{
+	  if(verbose)
+	      D (("Matched user: %s", s_user));
+      retval = -1; //We found at least one line for the user
+	  do
+	    {
+	      s_token = strtok_r (NULL, ":", &saveptr);
+	      if(verbose)
+		  D (("Authorization token: %s", s_token));
+	      if (s_token && strcmp (otp_id, s_token) == 0)
+		{
+		  if(verbose)
+		      D (("Match user/token as %s/%s", username, otp_id));
+		  fclose (opwfile);
+		  return 1;
+		}
+	    }
+	  while (s_token != NULL);
+	}
+    }
+
+  fclose (opwfile);
+
+  return retval;
+}
+
 #if HAVE_CR
 /* Fill buf with len bytes of random data */
 int generate_random(void *buf, int len)
@@ -163,7 +257,7 @@ int challenge_response(YK_KEY *yk, int slot,
 	memset(response, 0, res_size);
 
 	if (verbose) {
-		fprintf(stderr, "Sending %i bytes %s challenge to slot %i\n", len, (hmac == true)?"HMAC":"Yubico", slot);
+		fprintf(stderr, "Sending %u bytes %s challenge to slot %i\n", len, (hmac == true)?"HMAC":"Yubico", slot);
 		//_yk_hexdump(challenge, len);
 	}
 
@@ -216,7 +310,7 @@ get_user_challenge_file(YK_KEY *yk, const char *chalresp_path, const char *usern
     /* 0xffffffff == 4294967295 == 10 digits */
     len = strlen(chalresp_path == NULL ? "challenge" : username) + 1 + 10 + 1;
     if ((filename = malloc(len)) != NULL) {
-      int res = snprintf(filename, len, "%s-%i", chalresp_path == NULL ? "challenge" : username, serial);
+      int res = snprintf(filename, len, "%s-%u", chalresp_path == NULL ? "challenge" : username, serial);
       filename_malloced = 1;
       if (res < 0 || res > len) {
 	/* Not enough space, strangely enough. */
@@ -357,7 +451,7 @@ write_chalresp_state(FILE *f, CR_STATE *state)
   if (ftruncate(fd, 0))
     goto out;
 
-  fprintf(f, "v2:%s:%s:%s:%d:%d\n", challenge_hex, hashed_hex, salt_hex, iterations, state->slot);
+  fprintf(f, "v2:%s:%s:%s:%u:%d\n", challenge_hex, hashed_hex, salt_hex, iterations, state->slot);
 
   if (fflush(f) < 0)
     goto out;
@@ -370,3 +464,44 @@ write_chalresp_state(FILE *f, CR_STATE *state)
   return 0;
 }
 #endif /* HAVE_CR */
+
+size_t filter_result_len(const char *filter, const char *user, char *output) {
+  const char *part = NULL;
+  size_t result = 0;
+  do
+    {
+      size_t len;
+      part = strstr(filter, "%u");
+      if(part)
+        len = part - filter;
+      else
+        len = strlen(filter);
+      if (output)
+        {
+          strncpy(output, filter, len);
+          output += len;
+        }
+      result += len;
+      filter += len + 2;
+      if(part)
+        {
+          if(output)
+            {
+              strncpy(output, user, strlen(user));
+              output += strlen(user);
+            }
+          result += strlen(user);
+        }
+    }
+  while(part);
+
+  if(output)
+    *output = '\0';
+  return(result + 1);
+}
+
+char *filter_printf(const char *filter, const char *user) {
+  char *result = malloc(filter_result_len(filter, user, NULL));
+  filter_result_len(filter, user, result);
+  return result;
+}
