@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2011-2013 Yubico AB
+ * Copyright (c) 2011-2014 Yubico AB
  * Copyright (c) 2011 Tollef Fog Heen <tfheen@err.no>
  * All rights reserved.
  *
@@ -35,9 +35,7 @@
 #define __PAM_YUBICO_UTIL_H_INCLUDED__
 
 #include <stdio.h>
-
-#include <ykclient.h>
-
+#include <stdint.h>
 
 #if defined(DEBUG_PAM)
 # if defined(HAVE_SECURITY__PAM_MACROS_H)
@@ -55,12 +53,10 @@
 #endif /* DEBUG_PAM */
 
 int get_user_cfgfile_path(const char *common_path, const char *filename, const char *username, char **fn);
+int check_user_token(const char *authfile, const char *username, const char *otp_id, int verbose);
 
 #if HAVE_CR
-
 #include <ykcore.h>
-#include <ykstatus.h>
-#include <ykdef.h>
 
 /* Challenges can be 0..63 or 64 bytes long, depending on YubiKey configuration.
  * We settle for 63 bytes to have something that works with all configurations.
@@ -100,4 +96,7 @@ int challenge_response(YK_KEY *yk, int slot,
 
 #endif /* HAVE_CR */
 
+size_t filter_result_len(const char *filter, const char *user, char *output);
+char *filter_printf(const char *filter, const char *user);
+
 #endif /* __PAM_YUBICO_UTIL_H_INCLUDED__ */
